@@ -19,7 +19,10 @@
 
 #include "mdbtools.h"
 
-#define MDB_DEBUG_USAGE 0
+#ifdef DMALLOC
+#include "dmalloc.h"
+#endif
+
 
 static gint mdb_col_comparer(MdbColumn *a, MdbColumn *b)
 {
@@ -40,44 +43,61 @@ unsigned char mdb_col_needs_size(int col_type)
 	}
 }
 
-MdbTableDef *mdb_read_table(MdbCatalogEntry *entry)
+MdbTableDef *
+mdb_read_table(MdbCatalogEntry *entry)
 {
-MdbTableDef *table;
-MdbHandle *mdb = entry->mdb;
-MdbFormatConstants *fmt = mdb->fmt;
-int len;
-int rownum, row_start, row_end;
+	MdbTableDef *table;
+	MdbHandle *mdb = entry->mdb;
+	MdbFormatConstants *fmt = mdb->fmt;
+	int len;
+	int rownum, row_start, row_end;
+	guint32 pg;
 
 	table = mdb_alloc_tabledef(entry);
 
 	mdb_read_pg(mdb, entry->table_pg);
-	len = mdb_get_int16(mdb,8);
+	if (mdb->pg_buf[0] != 0x02) return NULL; /* not a valid table def page */
 
-	table->num_rows = mdb_get_int32(mdb, fmt->tab_num_rows_offset);
-	table->num_cols = mdb_get_int16(mdb, fmt->tab_num_cols_offset);
-	table->num_idxs = mdb_get_int32(mdb, fmt->tab_num_idxs_offset); 
-	table->num_real_idxs = mdb_get_int32(mdb, fmt->tab_num_ridxs_offset); 
+	len = mdb_pg_get_int16(mdb,8);
+
+	table->num_rows = mdb_pg_get_int32(mdb, fmt->tab_num_rows_offset);
+	table->num_cols = mdb_pg_get_int16(mdb, fmt->tab_num_cols_offset);
+	table->num_idxs = mdb_pg_get_int32(mdb, fmt->tab_num_idxs_offset); 
+	table->num_real_idxs = mdb_pg_get_int32(mdb, fmt->tab_num_ridxs_offset); 
 
 	/* grab a copy of the usage map */
 	rownum = mdb->pg_buf[fmt->tab_usage_map_offset];
-	mdb_read_alt_pg(mdb, mdb_get_int24(mdb, fmt->tab_usage_map_offset + 1)); 
+	pg = mdb_pg_get_int24(mdb, fmt->tab_usage_map_offset + 1); 
+	mdb_read_alt_pg(mdb, pg);
 	mdb_swap_pgbuf(mdb);
-	row_start = mdb_get_int16(mdb, (fmt->row_count_offset + 2) + (rownum*2));
-   row_end = mdb_find_end_of_row(mdb, rownum);
+	row_start = mdb_pg_get_int16(mdb, (fmt->row_count_offset + 2) + (rownum*2));
+	row_end = mdb_find_end_of_row(mdb, rownum);
 	table->map_sz = row_end - row_start + 1;
 	table->usage_map = malloc(table->map_sz);
 	memcpy(table->usage_map, &mdb->pg_buf[row_start], table->map_sz);
-#if MDB_DEBUG_USAGE
-	buffer_dump(mdb->pg_buf, row_start, row_end);
-#endif
+	if (mdb_get_option(MDB_DEBUG_USAGE)) 
+		buffer_dump(mdb->pg_buf, row_start, row_end);
 	/* swap back */
 	mdb_swap_pgbuf(mdb);
-#if MDB_DEBUG_USAGE
-	printf ("usage map found on page %ld start %d end %d\n", mdb_get_int24(mdb, fmt->tab_usage_map_offset + 1), row_start, row_end);
+	mdb_debug(MDB_DEBUG_USAGE,"usage map found on page %ld rownum %d start %d end %d", mdb_pg_get_int24(mdb, fmt->tab_usage_map_offset + 1), rownum, row_start, row_end);
+
+
+	/* now grab the free space page map */
+#if 1
+	//mdb_swap_pgbuf(mdb);
+	rownum = mdb->pg_buf[fmt->tab_free_map_offset];
+	mdb_read_alt_pg(mdb, mdb_pg_get_int24(mdb, fmt->tab_free_map_offset + 1));
+	mdb_swap_pgbuf(mdb);
+	row_start = mdb_pg_get_int16(mdb, (fmt->row_count_offset + 2) + (rownum*2));
+	row_end = mdb_find_end_of_row(mdb, rownum);
+	table->freemap_sz = row_end - row_start + 1;
+	table->free_usage_map = malloc(table->freemap_sz);
+	memcpy(table->free_usage_map, &mdb->pg_buf[row_start], table->freemap_sz);
+	mdb_swap_pgbuf(mdb);
 #endif
+	mdb_debug(MDB_DEBUG_USAGE,"free map found on page %ld rownum %d start %d end %d\n", mdb_pg_get_int24(mdb, fmt->tab_free_map_offset + 1), rownum, row_start, row_end);
 
-
-	table->first_data_pg = mdb_get_int16(mdb, fmt->tab_first_dpg_offset);
+	table->first_data_pg = mdb_pg_get_int16(mdb, fmt->tab_first_dpg_offset);
 
 	return table;
 }
@@ -86,26 +106,68 @@ int rownum, row_start, row_end;
 ** read the next page if offset is > pg_size
 ** return true if page was read
 */ 
-static int read_pg_if(MdbHandle *mdb, int *cur_pos, int offset)
+int 
+read_pg_if(MdbHandle *mdb, int *cur_pos, int offset)
 {
 	if (*cur_pos + offset >= mdb->fmt->pg_size) {
-		mdb_read_pg(mdb, mdb_get_int32(mdb,4));
+		mdb_read_pg(mdb, mdb_pg_get_int32(mdb,4));
 		*cur_pos = 8 - (mdb->fmt->pg_size - (*cur_pos));
 		return 1;
 	}
 	return 0;
 }
+guint32 
+read_pg_if_32(MdbHandle *mdb, int *cur_pos)
+{
+	unsigned char c[4];
+	int i, rc = 0;
+
+	for (i=0;i<4;i++) {
+		rc += read_pg_if(mdb, cur_pos, i);
+		c[i] = mdb->pg_buf[(*cur_pos) + i];
+	}
+	return mdb_get_int32(c, 0);
+}
+guint16 
+read_pg_if_16(MdbHandle *mdb, int *cur_pos)
+{
+	unsigned char low_byte, high_byte;
+	int rc = 0;
+
+	rc += read_pg_if(mdb, cur_pos, 0);
+	low_byte = mdb->pg_buf[*cur_pos];
+	rc += read_pg_if(mdb, cur_pos, 1);
+	high_byte = mdb->pg_buf[(*cur_pos) + 1];
+
+	return (high_byte * 256 + low_byte);
+}
+guint16 
+read_pg_if_n(MdbHandle *mdb, unsigned char *buf, int *cur_pos, int len)
+{
+	int half;
+
+	if (*cur_pos + len < mdb->fmt->pg_size) {
+		memcpy(buf, &mdb->pg_buf[*cur_pos], len);
+		return 0;
+	}
+	half = (mdb->fmt->pg_size - *cur_pos - 1);
+	memcpy(buf, &mdb->pg_buf[*cur_pos], half);
+	mdb_read_pg(mdb, mdb_pg_get_int32(mdb,4));
+	memcpy(buf, &mdb->pg_buf[8], len - half);
+	*cur_pos = 8 - (mdb->fmt->pg_size - (*cur_pos));
+	return 1;
+}
 
 GPtrArray *mdb_read_columns(MdbTableDef *table)
 {
-MdbHandle *mdb = table->entry->mdb;
-MdbFormatConstants *fmt = mdb->fmt;
-MdbColumn col, *pcol;
-int len, i,j;
-unsigned char low_byte, high_byte;
-int cur_col, cur_name;
-int name_sz;
-GSList	*slist = NULL;
+	MdbHandle *mdb = table->entry->mdb;
+	MdbFormatConstants *fmt = mdb->fmt;
+	MdbColumn col, *pcol;
+	int len, i,j;
+	unsigned char low_byte, high_byte;
+	int cur_col, cur_name;
+	int name_sz;
+	GSList	*slist = NULL;
 	
 	table->columns = g_ptr_array_new();
 
@@ -123,23 +185,28 @@ GSList	*slist = NULL;
 	buffer_dump(mdb->pg_buf, cur_col ,cur_col + 18); */
 #endif
 		memset(&col, 0, sizeof(col));
-		col.col_num = mdb->pg_buf[cur_col + fmt->col_num_offset];
 
 		read_pg_if(mdb, &cur_col, 0);
 		col.col_type = mdb->pg_buf[cur_col];
 
+		read_pg_if(mdb, &cur_col, fmt->col_num_offset); // col_num_offset == 1 or 5
+		col.col_num = mdb->pg_buf[cur_col + fmt->col_num_offset];
+
+		/* FIXME: can this be right in Jet3 and Jet4? */
 		if (col.col_type == MDB_NUMERIC) {
+			read_pg_if(mdb, &cur_col, 11);
 			col.col_prec = mdb->pg_buf[cur_col + 11];
+			read_pg_if(mdb, &cur_col, 12);
 			col.col_scale = mdb->pg_buf[cur_col + 12];
 		}
 
-		read_pg_if(mdb, &cur_col, 13);
+		read_pg_if(mdb, &cur_col, fmt->col_fixed_offset); // col_fixed_offset == 13 or 15
 		col.is_fixed = mdb->pg_buf[cur_col + fmt->col_fixed_offset] & 
-			0x01 ? 1 : 0;
+				0x01 ? 1 : 0;
 		if (col.col_type != MDB_BOOL) {
-			read_pg_if(mdb, &cur_col, 17);
+			read_pg_if(mdb, &cur_col, fmt->col_size_offset); // col_size_offset == 16 or 23
 			low_byte = mdb->pg_buf[cur_col + fmt->col_size_offset];
-			read_pg_if(mdb, &cur_col, 18);
+			read_pg_if(mdb, &cur_col, fmt->col_size_offset + 1);
 			high_byte = mdb->pg_buf[cur_col + fmt->col_size_offset + 1];
 			col.col_size += high_byte * 256 + low_byte;
 		} else
@@ -178,15 +245,15 @@ GSList	*slist = NULL;
 			/* name wrapped over page */
 			if (len < name_sz) {
 				/* read the next pg */
-				mdb_read_pg(mdb, mdb_get_int32(mdb,4)); 
+				mdb_read_pg(mdb, mdb_pg_get_int32(mdb,4)); 
 				cur_name = 8 - (fmt->pg_size - cur_name);
-				if (len % 2) cur_name++;
 				/* get the rest of the name */
-				for (j=0;j<len;j+=2) {
+				for (j=len;j<name_sz;j++) {
+					if (!(j%2))
+						pcol->name[j/2]	= mdb->pg_buf[cur_name + j];
 				}
-				memcpy(&pcol->name[len], &mdb->pg_buf[cur_name], name_sz - len);
 			}
-			pcol->name[name_sz]='\0';
+			pcol->name[name_sz/2]='\0';
 
 			cur_name += name_sz;
 		} else if (IS_JET3(mdb)) {
@@ -201,7 +268,7 @@ GSList	*slist = NULL;
 			/* name wrapped over page */
 			if (len < name_sz) {
 				/* read the next pg */
-				mdb_read_pg(mdb, mdb_get_int32(mdb,4)); 
+				mdb_read_pg(mdb, mdb_pg_get_int32(mdb,4)); 
 				cur_name = 8 - (fmt->pg_size - cur_name);
 				/* get the rest of the name */
 				memcpy(&pcol->name[len], &mdb->pg_buf[cur_name], name_sz - len);
@@ -233,14 +300,13 @@ int coln;
 MdbIndex *idx;
 MdbHandle *mdb = entry->mdb;
 int i,bitn;
-int pgnum;
+guint32 pgnum;
 
 	table = mdb_read_table(entry);
 	fprintf(stdout,"definition page     = %lu\n",entry->table_pg);
 	fprintf(stdout,"number of datarows  = %d\n",table->num_rows);
 	fprintf(stdout,"number of columns   = %d\n",table->num_cols);
 	fprintf(stdout,"number of indices   = %d\n",table->num_real_idxs);
-	fprintf(stdout,"first data page     = %lu\n",table->first_data_pg);
 
 	mdb_read_columns(table);
 	mdb_read_indices(table);
@@ -259,15 +325,17 @@ int pgnum;
 		mdb_index_dump(table, idx);
 	}
 	if (table->usage_map) {
-		printf("pages reserved by this object\n",pgnum);
-		pgnum = _mdb_get_int32(table->usage_map,1);
+		printf("pages reserved by this object\n");
+		printf("usage map pg %lu\n", table->map_base_pg);
+		printf("free map pg %lu\n", table->freemap_base_pg);
+		pgnum = mdb_get_int32(table->usage_map,1);
 		/* the first 5 bytes of the usage map mean something */
 		coln = 0;
 		for (i=5;i<table->map_sz;i++) {
 			for (bitn=0;bitn<8;bitn++) {
 				if (table->usage_map[i] & 1 << bitn) {
 					coln++;
-					printf("%6ld ",pgnum);
+					printf("%6lu ",(long unsigned) pgnum);
 					if (coln==10) {
 						printf("\n");
 						coln = 0;
@@ -276,5 +344,6 @@ int pgnum;
 				pgnum++;
 			}
 		}
+		printf("\n");
 	}
 }
