@@ -32,7 +32,7 @@
 
 #include "connectparams.h"
 
-static char  software_version[]   = "$Id: odbc.c,v 1.9 2003/01/15 02:02:09 brianb Exp $";
+static char  software_version[]   = "$Id: odbc.c,v 1.12 2004/02/08 21:54:20 brianb Exp $";
 static void *no_unused_var_warn[] = {software_version,
                                      no_unused_var_warn};
 
@@ -47,7 +47,11 @@ static SQLRETURN SQL_API _SQLFreeConnect(SQLHDBC hdbc);
 static SQLRETURN SQL_API _SQLFreeEnv(SQLHENV henv);
 static SQLRETURN SQL_API _SQLFreeStmt(SQLHSTMT hstmt, SQLUSMALLINT fOption);
 
+static void bind_columns (struct _hstmt*);
+
+#ifndef MIN
 #define MIN(a,b) (a>b ? b : a)
+#endif
 #define _MAX_ERROR_LEN 255
 static char lastError[_MAX_ERROR_LEN+1];
 
@@ -785,10 +789,10 @@ struct _sql_bind_info *cur;
         while (cur) {
             if (cur->column_number>0 &&
             cur->column_number <= env->sql->num_columns) {
-				mdbsql_bind_column(env->sql, 
+				mdb_sql_bind_column(env->sql, 
 					cur->column_number, cur->varaddr);
 				if (cur->column_lenbind)
-					mdbsql_bind_len(env->sql, 
+					mdb_sql_bind_len(env->sql, 
 						cur->column_number, cur->column_lenbind);
             } else {
                 /* log error ? */
@@ -1058,7 +1062,7 @@ int i;
 	}
 	
 	strcpy(rgbValue,
-		mdb_col_to_string(mdb, col->cur_value_start, col->col_type, 
+		mdb_col_to_string(mdb, mdb->pg_buf, col->cur_value_start, col->col_type, 
 		col->cur_value_len));
 	//*((char *)&rgbValue[col->cur_value_len])='\0';
 	*pcbValue = col->cur_value_len;
@@ -1445,6 +1449,8 @@ static int _odbc_get_server_type(int clt_type)
 	case SQL_DECIMAL:
 	case SQL_NUMERIC:
 	case SQL_FLOAT:
+	default:
+		break;
 	}
 }
 static SQLSMALLINT _odbc_get_client_type(int srv_type)
