@@ -20,15 +20,10 @@
 
 #include "mdbtools.h"
 
+int
 main(int argc, char **argv)
 {
-int rows;
-int i;
-unsigned char buf[2048];
 MdbHandle *mdb;
-MdbCatalogEntry entry;
-MdbTableDef *table;
-GList *l;
 int j;
 int page, start, stop;
 
